/*
m_formant - A formant filter

DESCRIPTION
m_formant is a formant filter, imprinting vowel type characteristics on an
audio input. It allows to create all unique three vowel "words" (see the
kvowel parameter).

SYNTAX
aout m_formant knote, kkeytrack, ain, kresonance, kregister, kvowel, ifnfreq, \
		ifnamp, ifnbw, ifntrack, ifnvowel

INITIALIZATION
ifnfreq, ifnamp, ifnbw - Function tables holding the coefficients for the
		formants
ifntrack - Function table holding the indexes for keytracking
ifnvowel - Function table defining the succession of vowels.

PERFORMANCE
knote - MIDI notenumber (0-127), this is a k-rate parameter to allow for
		portamento
kkeytrack - Keytracking of the formant/vowel of the filter. 0 is no keytracking
		1 is normal (100%) keytracking and -1 is -100%, i.e. reverted keyboard
ain - Audio input signal
kresonance - Resonance/mixing of the vocal filter 0-1
kregister - Vocal register or type (bass, tenor, countertenor, alto and
		soprano) between 0-4
kvowel - Formant/vowel of the filter 0-31
		a e i o u a e o u e i a o u i e o a i u e o i a u e a o i u a o

CREDITS
Author: Jeanette C., thanks to the Csound developers and contributors to
make this possible!
*/

opcode m_formant, a, kkakkkiiiii
	knote, kkeytrack, ain, kresonance, kregister, kvowel, ifnfreq, ifnamp, \
	ifnbw, ifntrack, ifnvowel xin

	; Only process the audio if resonance not 0
	if (kresonance != 0) then
		; Find register index
		ktmpreg = kregister + (knote * kkeytrack)
		krealreg limit ktmpreg, 1, 128
		
		; Find vowel indices
		;ktmpvowel = (kvowel / 8) + 1
		ktmpvowel = kvowel
		klovowel = floor(ktmpvowel)
		khivowel = klovowel + 1
		kfracvowel = frac(ktmpvowel)
		
		; Retrieve register and vowel lookup values
		ktruereg table krealreg, ifntrack, 0, 0, 0
		klotruereg = floor(ktruereg)
		khitruereg = klotruereg + 1
		kfractruereg = frac(ktruereg)

		klotruevowel table klovowel, ifnvowel, 0, 0, 0
		khitruevowel table khivowel, ifnvowel, 0, 0, 0
		
		; Retrieve low vowel from low register
		kllvowel1 table (klotruevowel * 25 + klotruereg * 5)+1, ifnfreq, 0, 0, 0
		kllvowel2 table (klotruevowel * 25 + klotruereg * 5)+2, ifnfreq, 0, 0, 0
		kllvowel3 table (klotruevowel * 25 + klotruereg * 5)+3, ifnfreq, 0, 0, 0
		kllvowel4 table (klotruevowel * 25 + klotruereg * 5)+4, ifnfreq, 0, 0, 0
		kllvowel5 table (klotruevowel * 25 + klotruereg * 5)+5, ifnfreq, 0, 0, 0
		kllamp1 table (klotruevowel * 25 + klotruereg * 5)+1, ifnamp, 0, 0, 0
		kllamp2 table (klotruevowel * 25 + klotruereg * 5)+2, ifnamp, 0, 0, 0
		kllamp3 table (klotruevowel * 25 + klotruereg * 5)+3, ifnamp, 0, 0, 0
		kllamp4 table (klotruevowel * 25 + klotruereg * 5)+4, ifnamp, 0, 0, 0
		kllamp5 table (klotruevowel * 25 + klotruereg * 5)+5, ifnamp, 0, 0, 0
		kllbw1 table (klotruevowel * 25 + klotruereg * 5)+1, ifnbw, 0, 0, 0
		kllbw2 table (klotruevowel * 25 + klotruereg * 5)+2, ifnbw, 0, 0, 0
		kllbw3 table (klotruevowel * 25 + klotruereg * 5)+3, ifnbw, 0, 0, 0
		kllbw4 table (klotruevowel * 25 + klotruereg * 5)+4, ifnbw, 0, 0, 0
		kllbw5 table (klotruevowel * 25 + klotruereg * 5)+5, ifnbw, 0, 0, 0

		; Retrieve low register, high vowel
		klhvowel1 table (khitruevowel * 25 + klotruereg * 5)+1, ifnfreq, 0, 0, 0
		klhvowel2 table (khitruevowel * 25 + klotruereg * 5)+2, ifnfreq, 0, 0, 0
		klhvowel3 table (khitruevowel * 25 + klotruereg * 5)+3, ifnfreq, 0, 0, 0
		klhvowel4 table (khitruevowel * 25 + klotruereg * 5)+4, ifnfreq, 0, 0, 0
		klhvowel5 table (khitruevowel * 25 + klotruereg * 5)+5, ifnfreq, 0, 0, 0
		klhamp1 table (khitruevowel * 25 + klotruereg * 5)+1, ifnamp, 0, 0, 0
		klhamp2 table (khitruevowel * 25 + klotruereg * 5)+2, ifnamp, 0, 0, 0
		klhamp3 table (khitruevowel * 25 + klotruereg * 5)+3, ifnamp, 0, 0, 0
		klhamp4 table (khitruevowel * 25 + klotruereg * 5)+4, ifnamp, 0, 0, 0
		klhamp5 table (khitruevowel * 25 + klotruereg * 5)+5, ifnamp, 0, 0, 0
		klhbw1 table (khitruevowel * 25 + klotruereg * 5)+1, ifnbw, 0, 0, 0
		klhbw2 table (khitruevowel * 25 + klotruereg * 5)+2, ifnbw, 0, 0, 0
		klhbw3 table (khitruevowel * 25 + klotruereg * 5)+3, ifnbw, 0, 0, 0
		klhbw4 table (khitruevowel * 25 + klotruereg * 5)+4, ifnbw, 0, 0, 0
		klhbw5 table (khitruevowel * 25 + klotruereg * 5)+5, ifnbw, 0, 0, 0

		; Retrieve high register, low vowel
		khlvowel1 table (klotruevowel * 25 + khitruereg * 5)+1, ifnfreq, 0, 0, 0
		khlvowel2 table (klotruevowel * 25 + khitruereg * 5)+2, ifnfreq, 0, 0, 0
		khlvowel3 table (klotruevowel * 25 + khitruereg * 5)+3, ifnfreq, 0, 0, 0
		khlvowel4 table (klotruevowel * 25 + khitruereg * 5)+4, ifnfreq, 0, 0, 0
		khlvowel5 table (klotruevowel * 25 + khitruereg * 5)+5, ifnfreq, 0, 0, 0
		khlamp1 table (klotruevowel * 25 + khitruereg * 5)+1, ifnamp, 0, 0, 0
		khlamp2 table (klotruevowel * 25 + khitruereg * 5)+2, ifnamp, 0, 0, 0
		khlamp3 table (klotruevowel * 25 + khitruereg * 5)+3, ifnamp, 0, 0, 0
		khlamp4 table (klotruevowel * 25 + khitruereg * 5)+4, ifnamp, 0, 0, 0
		khlamp5 table (klotruevowel * 25 + khitruereg * 5)+5, ifnamp, 0, 0, 0
		khlbw1 table (klotruevowel * 25 + khitruereg * 5)+1, ifnbw, 0, 0, 0
		khlbw2 table (klotruevowel * 25 + khitruereg * 5)+2, ifnbw, 0, 0, 0
		khlbw3 table (klotruevowel * 25 + khitruereg * 5)+3, ifnbw, 0, 0, 0
		khlbw4 table (klotruevowel * 25 + khitruereg * 5)+4, ifnbw, 0, 0, 0
		khlbw5 table (klotruevowel * 25 + khitruereg * 5)+5, ifnbw, 0, 0, 0

		; Retrieve high register, high vowel
		khhvowel1 table (khitruevowel * 25 + khitruereg * 5)+1, ifnfreq, 0, 0, 0
		khhvowel2 table (khitruevowel * 25 + khitruereg * 5)+2, ifnfreq, 0, 0, 0
		khhvowel3 table (khitruevowel * 25 + khitruereg * 5)+3, ifnfreq, 0, 0, 0
		khhvowel4 table (khitruevowel * 25 + khitruereg * 5)+4, ifnfreq, 0, 0, 0
		khhvowel5 table (khitruevowel * 25 + khitruereg * 5)+5, ifnfreq, 0, 0, 0
		khhamp1 table (khitruevowel * 25 + khitruereg * 5)+1, ifnamp, 0, 0, 0
		khhamp2 table (khitruevowel * 25 + khitruereg * 5)+2, ifnamp, 0, 0, 0
		khhamp3 table (khitruevowel * 25 + khitruereg * 5)+3, ifnamp, 0, 0, 0
		khhamp4 table (khitruevowel * 25 + khitruereg * 5)+4, ifnamp, 0, 0, 0
		khhamp5 table (khitruevowel * 25 + khitruereg * 5)+5, ifnamp, 0, 0, 0
		khhbw1 table (khitruevowel * 25 + khitruereg * 5)+1, ifnamp, 0, 0, 0
		khhbw2 table (khitruevowel * 25 + khitruereg * 5)+2, ifnamp, 0, 0, 0
		khhbw3 table (khitruevowel * 25 + khitruereg * 5)+3, ifnamp, 0, 0, 0
		khhbw4 table (khitruevowel * 25 + khitruereg * 5)+4, ifnamp, 0, 0, 0
		khhbw5 table (khitruevowel * 25 + khitruereg * 5)+5, ifnamp, 0, 0, 0

		; Add the retrieved values with proper weight (use the fractional
		; register and vowel numbers for that)
		; final_value = (start_value + fractional_weight * (end_value - start_value)
		; first weight values in the low register, then in the high register
		; and then between the low and high register.
		klvowel1 = (kllvowel1 + kfracvowel * (klhvowel1 - kllvowel1))
		khvowel1 = (khlvowel1 + kfracvowel * (khhvowel1 - khlvowel1))
		kvowel1 = (klvowel1 + kfractruereg * (khvowel1 - klvowel1))
		klvowel2 = (kllvowel2 + kfracvowel * (klhvowel2 - kllvowel2))
		khvowel2 = (khlvowel2 + kfracvowel * (khhvowel2 - khlvowel2))
		kvowel2 = (klvowel2 + kfractruereg * (khvowel2 - klvowel2))
		klvowel3 = (kllvowel3 + kfracvowel * (klhvowel3 - kllvowel3))
		khvowel3 = (khlvowel3 + kfracvowel * (khhvowel3 - khlvowel3))
		kvowel3 = (klvowel3 + kfractruereg * (khvowel3 - klvowel3))
		klvowel4 = (kllvowel4 + kfracvowel * (klhvowel4 - kllvowel4))
		khvowel4 = (khlvowel4 + kfracvowel * (khhvowel4 - khlvowel4))
		kvowel4 = (klvowel4 + kfractruereg * (khvowel4 - klvowel4))
		klvowel5 = (kllvowel5 + kfracvowel * (klhvowel5 - kllvowel5))
		khvowel5 = (khlvowel5 + kfracvowel * (khhvowel5 - khlvowel5))
		kvowel5 = (klvowel5 + kfractruereg * (khvowel5 - klvowel5))

		klamp1 = (kllamp1 + kfracvowel * (klhamp1 - kllamp1))
		khamp1 = (khlamp1 + kfracvowel * (khhamp1 - khlamp1))
		kamp1 = (klamp1 + kfractruereg * (khamp1 - klamp1))
		klamp2 = (kllamp2 + kfracvowel * (klhamp2 - kllamp2))
		khamp2 = (khlamp2 + kfracvowel * (khhamp2 - khlamp2))
		kamp2 = (klamp2 + kfractruereg * (khamp2 - klamp2))
		klamp3 = (kllamp3 + kfracvowel * (klhamp3 - kllamp3))
		khamp3 = (khlamp3 + kfracvowel * (khhamp3 - khlamp3))
		kamp3 = (klamp3 + kfractruereg * (khamp3 - klamp3))
		klamp4 = (kllamp4 + kfracvowel * (klhamp4 - kllamp4))
		khamp4 = (khlamp4 + kfracvowel * (khhamp4 - khlamp4))
		kamp4 = (klamp4 + kfractruereg * (khamp4 - klamp4))
		klamp5 = (kllamp5 + kfracvowel * (klhamp5 - kllamp5))
		khamp5 = (khlamp5 + kfracvowel * (khhamp5 - khlamp5))
		kamp5 = (klamp5 + kfractruereg * (khamp5 - klamp5))

		klbw1 = (kllbw1 + kfracvowel * (klhbw1 - kllbw1))
		khbw1 = (khlbw1 + kfracvowel * (khhbw1 - khlbw1))
		kbw1 = (klbw1 + kfractruereg * (khbw1 - klbw1))
		klbw2 = (kllbw2 + kfracvowel * (klhbw2 - kllbw2))
		khbw2 = (khlbw2 + kfracvowel * (khhbw2 - khlbw2))
		kbw2 = (klbw2 + kfractruereg * (khbw2 - klbw2))
		klbw3 = (kllbw3 + kfracvowel * (klhbw3 - kllbw3))
		khbw3 = (khlbw3 + kfracvowel * (khhbw3 - khlbw3))
		kbw3 = (klbw3 + kfractruereg * (khbw3 - klbw3))
		klbw4 = (kllbw4 + kfracvowel * (klhbw4 - kllbw4))
		khbw4 = (khlbw4 + kfracvowel * (khhbw4 - khlbw4))
		kbw4 = (klbw4 + kfractruereg * (khbw4 - klbw4))
		klbw5 = (kllbw5 + kfracvowel * (klhbw5 - kllbw5))
		khbw5 = (khlbw5 + kfracvowel * (khhbw5 - khlbw5))
		kbw5 = (klbw5 + kfractruereg * (khbw5 - klbw5))

		; Apply the filters
		a1 butbp ain*db(kamp1), kvowel1, kbw1
		a2 butbp ain*db(kamp2), kvowel2, kbw2
		a3 butbp ain*db(kamp3), kvowel3, kbw3
		a4 butbp ain*db(kamp4), kvowel4, kbw4
		a5 butbp ain*db(kamp5), kvowel5, kbw5

		; Create the audio output of the filter
		aout = ain * (1-kresonance) + kresonance * (a1 + a2 + a3 + a4 + a5)
		goto Output
	else
		aout = ain
		goto Output
	endif

Output:
	xout aout
endop
